
from enum import IntEnum
from math import ceil
from typing import Iterator
from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QPainterPath
from PyQt5.QtWidgets import QGraphicsItem

from .init_values import (
    canvas,
    options,
    BoxLayoutMode,
    PortMode,
    PortType,
    PortSubType,
    BoxType)

from .utils import get_portgroup_name_from_ports_names
from .box_widget_moth import BoxWidgetMoth, UnwrapButton, TitleLine


def list_port_types_and_subs() -> Iterator[tuple[PortType, PortSubType]]:
    ''' simple fast generator port PortType and PortSubType preventing
        incoherent couples '''
    for port_type in PortType:
        if port_type is PortType.NULL:
            continue
        
        for port_subtype in PortSubType:
            if ((port_subtype is PortSubType.A2J
                    and not port_type is PortType.MIDI_JACK)
                or (port_subtype is PortSubType.CV
                    and not port_type is PortType.AUDIO_JACK)):
                # No such port should exist, it is just for win some time
                continue
            
            yield (port_type, port_subtype)


class TitleOn(IntEnum):
    TOP = 0
    SIDE = 1
    SIDE_UNDER_ICON = 2


class BoxWidget(BoxWidgetMoth):
    def __init__(self, group_id: int, group_name: str,
                 icon_type: int, icon_name: str):
        BoxWidgetMoth.__init__(
            self, group_id, group_name, icon_type, icon_name)
        self.update_positions_pending = False
        self._ex_width = self._width
        self._ex_height = self._height

        self._ex_scene_pos = self.scenePos()
        self._ex_ports_y_segments_dict = dict[str, list[list[int]]]()

    def __repr__(self) -> str:
        return f"BoxWidget({self._group_name}, {self._current_port_mode.name})"

    def _get_portgroup_name(self, portgrp_id: int):
        return get_portgroup_name_from_ports_names(
            [p.port_name for p in self._port_list
             if p.portgrp_id == portgrp_id])

    def _should_align_port_types(self) -> bool:
        ''' check if we can align port types
            eg, align first midi input to first midi output '''
        port_types_aligner = list[tuple[int, int]]()
        
        for port_type, port_subtype in list_port_types_and_subs():
            n_ins = 0
            n_outs = 0

            for port in self._port_list:
                if (port.port_type == port_type
                        and port.port_subtype == port_subtype):
                    if port.port_mode is PortMode.INPUT:
                        n_ins += 1
                    elif port.port_mode is PortMode.OUTPUT:
                        n_outs += 1

            port_types_aligner.append((n_ins, n_outs))

        winner = PortMode.NULL

        for n_ins, n_outs in port_types_aligner:
            if ((winner is PortMode.INPUT and n_outs > n_ins)
                    or (winner is PortMode.OUTPUT and n_ins > n_outs)):
                return False

            if n_ins > n_outs:
                winner = PortMode.INPUT
            elif n_outs > n_ins:
                winner = PortMode.OUTPUT
        
        return True
    
    def _get_geometry_dict(
            self, align_port_types: bool) -> dict:
        max_in_width = max_out_width = 0
        last_in_pos = last_out_pos = 0
        final_last_in_pos = final_last_out_pos = last_in_pos
        
        box_theme = self.get_theme()
        port_spacing = box_theme.port_spacing()
        port_in_offset = box_theme.port_in_offset()
        port_out_offset = box_theme.port_out_offset()
        port_type_spacing = box_theme.port_type_spacing()
        last_in_type_alter = (PortType.NULL, PortSubType.REGULAR)
        last_out_type_alter = (PortType.NULL, PortSubType.REGULAR)
        last_port_mode = PortMode.NULL
        
        for port_type, port_subtype in list_port_types_and_subs():                
            for port in self._port_list:
                if (port.port_type is not port_type
                        or port.port_subtype is not port_subtype):
                    continue

                last_of_portgrp = bool(port.pg_pos + 1 == port.pg_len)
                size = 0
                max_pwidth = options.max_port_width
    
                if port.port_mode is PortMode.INPUT:
                    port_offset = port_in_offset
                else:
                    port_offset = port_out_offset

                if port.portgrp_id:
                    portgrp = canvas.get_portgroup(self._group_id, port.portgrp_id)
                    if port.pg_pos == 0:
                        portgrp_name = self._get_portgroup_name(port.portgrp_id)

                        portgrp.widget.set_print_name(
                            portgrp_name,
                            max_pwidth - canvas.theme.port_grouped_width - 5)
                    
                    port.widget.set_print_name(
                        port.port_name.replace(
                            self._get_portgroup_name(port.portgrp_id), '', 1),
                        int(max_pwidth/2))
                    
                    if (portgrp.widget.get_text_width() + 5
                            > max_pwidth - port.widget.get_text_width()):
                        portgrp.widget.reduce_print_name(
                            max_pwidth - port.widget.get_text_width() - 5)
                    
                    # the port_grouped_width is also used to define
                    # the portgroup minimum width
                    size = (max(portgrp.widget.get_text_width(),
                                canvas.theme.port_grouped_width)
                            + max(port.widget.get_text_width() + 6,
                                    canvas.theme.port_grouped_width)
                            + port_offset)
                else:
                    port.widget.set_print_name(port.port_name, max_pwidth)
                    size = max(port.widget.get_text_width() + port_offset, 20)
                
                type_and_sub = (port.port_type, port.port_subtype)
                
                if port.port_mode is PortMode.INPUT:
                    max_in_width = max(max_in_width, size)
                    if type_and_sub != last_in_type_alter:
                        if last_in_type_alter != (PortType.NULL, PortSubType.REGULAR):
                            last_in_pos += port_type_spacing
                        last_in_type_alter = type_and_sub

                    last_in_pos += canvas.theme.port_height
                    if last_of_portgrp:
                        last_in_pos += port_spacing

                elif port.port_mode is PortMode.OUTPUT:
                    max_out_width = max(max_out_width, size)
                    
                    if type_and_sub != last_out_type_alter:
                        if last_out_type_alter != (PortType.NULL, PortSubType.REGULAR):
                            last_out_pos += port_type_spacing
                        last_out_type_alter = type_and_sub
                    
                    last_out_pos += canvas.theme.port_height
                    if last_of_portgrp:
                        last_out_pos += port_spacing
                
                final_last_in_pos = last_in_pos
                final_last_out_pos = last_out_pos
            
            if align_port_types:
                # align port types horizontally
                if last_in_pos > last_out_pos:
                    last_out_type_alter = last_in_type_alter
                else:
                    last_in_type_alter = last_out_type_alter
                last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)
        
        # calculates height in case of one column only
        last_inout_pos = 0
        last_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
        
        if self._current_port_mode is PortMode.BOTH:
            for port_type, port_subtype in list_port_types_and_subs():
                for port in self._port_list:
                    if (port.port_type is not port_type
                            or port.port_subtype is not port_subtype):
                        continue

                    if (port.port_type, port.port_subtype) != last_type_and_sub:
                        if last_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                            last_inout_pos += port_type_spacing
                        last_type_and_sub = (port.port_type, port.port_subtype)

                    if port.pg_pos:
                        continue

                    last_inout_pos += port.pg_len * canvas.theme.port_height
                    last_inout_pos += port_spacing
                    
                    last_port_mode = port.port_mode
        
        return {'last_in_pos': final_last_in_pos,
                'last_out_pos': final_last_out_pos,
                'last_inout_pos': last_inout_pos,
                'max_in_width': max_in_width,
                'max_out_width': max_out_width,
                'last_port_mode': last_port_mode}

    def _set_ports_y_positions(
            self, align_port_types: bool, start_pos: int, one_column: bool
                ) -> dict[str, list[list[int]]]:
        def set_widget_pos(widget: QGraphicsItem, pos: float):
            if self._wrapping:
                widget.setY(pos - ((pos - wrapped_port_pos)
                                   * self._wrapping_ratio))
            elif self._unwrapping:
                widget.setY(wrapped_port_pos + ((pos - wrapped_port_pos)
                                                * self._wrapping_ratio))
            elif self._wrapped:
                widget.setY(wrapped_port_pos)
            else:
                widget.setY(pos)
            
        ''' ports Y positioning, and get height segments info
            used if port-in-offset or port-out-offset are not zero in box theme'''
        last_in_pos = last_out_pos = start_pos
        wrapped_port_pos = start_pos
        
        box_theme = self.get_theme()
        port_spacing = box_theme.port_spacing()
        port_type_spacing = box_theme.port_type_spacing()
        last_in_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
        last_out_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
        last_type_and_sub = (PortType.NULL, PortSubType.REGULAR)
        input_segments = list[list[int]]()
        output_segments = list[list[int]]()
        in_segment = [last_in_pos, last_in_pos]
        out_segment = [last_out_pos, last_out_pos]
        
        for port_type, port_subtype in list_port_types_and_subs():                
            for port in self._port_list:
                if (port.port_type is not port_type
                        or port.port_subtype is not port_subtype):
                    continue
                
                if one_column:
                    last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)
                
                if port.portgrp_id and port.pg_pos > 0:
                    continue
                
                type_and_sub = (port.port_type, port.port_subtype)
                if one_column:
                    if type_and_sub != last_type_and_sub:
                        if last_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                            last_in_pos += port_type_spacing
                            last_out_pos += port_type_spacing
                        last_type_and_sub = type_and_sub
                
                if port.port_mode is PortMode.INPUT:
                    if not one_column and type_and_sub != last_in_type_and_sub:
                        if last_in_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                            last_in_pos += port_type_spacing
                        last_in_type_and_sub = type_and_sub
                    
                    if last_in_pos >= in_segment[1] + port_spacing + port_type_spacing:
                        if in_segment[0] != in_segment[1]:
                            input_segments.append(in_segment)
                        in_segment = [last_in_pos, last_in_pos]
                    
                    if port.portgrp_id:
                        # we place the portgroup widget and all its ports now
                        # because in one column mode, we can't be sure
                        # that port consecutivity isn't break by a port with
                        # another mode:
                        # 
                        # input L
                        #     output L
                        # input R
                        #     output R
                        portgrp = port.portgrp
                        if portgrp.widget is not None:
                            set_widget_pos(portgrp.widget, last_in_pos)
                        
                        for gp_port in portgrp.ports:
                            set_widget_pos(gp_port.widget, last_in_pos)
                            last_in_pos += canvas.theme.port_height
                    else:
                        set_widget_pos(port.widget, last_in_pos)
                        last_in_pos += canvas.theme.port_height
                    in_segment[1] = last_in_pos
                    last_in_pos += port_spacing

                elif port.port_mode is PortMode.OUTPUT:
                    if not one_column and type_and_sub != last_out_type_and_sub:
                        if last_out_type_and_sub != (PortType.NULL, PortSubType.REGULAR):
                            last_out_pos += port_type_spacing
                        last_out_type_and_sub = type_and_sub

                    if last_out_pos >= out_segment[1] + port_spacing + port_type_spacing:
                        if out_segment[0] != out_segment[1]:
                            output_segments.append(out_segment)
                        out_segment = [last_out_pos, last_out_pos]

                    if port.portgrp_id:
                        portgrp = port.portgrp
                        if portgrp.widget is not None:
                            set_widget_pos(portgrp.widget, last_out_pos)

                        for gp_port in portgrp.ports:
                            set_widget_pos(gp_port.widget, last_out_pos)
                            last_out_pos += canvas.theme.port_height
                    else:
                        set_widget_pos(port.widget, last_out_pos)
                        last_out_pos += canvas.theme.port_height
                    
                    out_segment[1] = last_out_pos
                    last_out_pos += port_spacing
            
            if align_port_types:
                # align port types horizontally
                if last_in_pos > last_out_pos:
                    last_out_type_and_sub = last_in_type_and_sub
                else:
                    last_in_type_and_sub = last_out_type_and_sub
                last_in_pos = last_out_pos = max(last_in_pos, last_out_pos)
        
        if in_segment[0] != in_segment[1]:
            input_segments.append(in_segment)
        if out_segment[0] != out_segment[1]:
            output_segments.append(out_segment)
        
        return {'input_segments': input_segments,
                'output_segments': output_segments}
    
    @staticmethod
    def split_in_two(string: str, n_lines: int) -> list[str]:
        def polished_list(input_list: list) -> list:
            output_list = list[str]()

            for string in input_list:
                if not string:
                    continue
                if len(string) == 1:
                    if output_list:
                        output_list[-1] += string
                    else:
                        output_list.append(string)
                else:
                    if output_list and len(output_list[-1]) <= 1:
                        output_list[-1] += string
                    else:
                        output_list.append(string)
            
            return [s.strip() for s in output_list]
                
        if n_lines <= 1:
            return [string]
            
        sep_indexes = list[int]()
        last_was_digit = False
        last_was_upper = False

        for i in range(len(string)):
            c = string[i]
            if c in (' ', '-', '_', '.'):
                sep_indexes.append(i)
            else:
                if c.upper() == c:
                    if c.isdigit():
                        if not last_was_digit:
                            sep_indexes.append(i)
                    else:
                        if last_was_digit or not last_was_upper:
                            sep_indexes.append(i)

                    last_was_upper = True
                else:
                    if last_was_digit:
                        sep_indexes.append(i)
                    last_was_upper = False

                last_was_digit = c.isdigit()
            
        if not sep_indexes:
            return [string]
        
        if len(sep_indexes) + 1 <= n_lines:
            return_list = list[str]()
            last_index = 0

            for sep_index in sep_indexes:
                return_list.append(string[last_index:sep_index])
                last_index = sep_index

            return_list.append(string[last_index:])

            return polished_list(return_list)
        
        best_indexes = [0]
        string_rest = string

        for i in range(n_lines, 1, -1):
            
            target = best_indexes[-1] + int(len(string_rest)/i)
            best_index = None
            best_dif = len(string)

            for s in sep_indexes:
                if s <= best_indexes[-1]:
                    continue

                dif = abs(target - s)
                if dif < best_dif:
                    best_index = s
                    best_dif = dif
                else:
                    break

            if best_index is None:
                continue

            string_rest = string[best_index:]
            best_indexes.append(best_index)

        best_indexes = best_indexes[1:]
        last_index = 0
        return_list = list[str]()

        for i in best_indexes:
            return_list.append(string[last_index:i])
            last_index = i

        return_list.append(string[last_index:])
        return polished_list(return_list)
    
    
    def _split_title(self, n_lines: int) -> tuple[TitleLine]:
        title, slash, subtitle = self._group_name.partition('/')

        if (not subtitle
                and self._box_type == BoxType.CLIENT
                and ' (' in self._group_name
                and self._group_name.endswith(')')):
            title, parenthese, subtitle = self._group_name.partition(' (')
            subtitle = subtitle[:-1]
        
        theme = self.get_theme()

        if self._box_type == BoxType.CLIENT and subtitle:
            # if there is a subtitle, title is not bold when subtitle is.
            # so title is 'little'
            client_line = TitleLine(title, theme, little=True)
            subclient_line = TitleLine(subtitle, theme)
            title_lines = []
            
            if n_lines <= 2:
                title_lines.append(client_line)
                title_lines.append(subclient_line)
            
            else:
                if client_line.size > subclient_line.size:
                    client_strs = self.split_in_two(title, 2)
                    for client_str in client_strs:
                        title_lines.append(TitleLine(client_str, theme, little=True))
                    
                    for subclient_str in self.split_in_two(subtitle, n_lines - 2):
                        title_lines.append(TitleLine(subclient_str, theme))
                else:
                    two_lines_title = False
                    
                    if n_lines >= 4:
                        # Check if we need to split the client title
                        # it could be "Carla-Multi-Client.Carla".
                        subtitles = self.split_in_two(subtitle, n_lines - 2)

                        for subtt in subtitles:
                            subtt_line = TitleLine(subtt, theme)
                            if subtt_line.size > client_line.size:
                                break
                        else:
                            client_strs = self.split_in_two(title, 2)
                            for client_str in client_strs:
                                title_lines.append(
                                    TitleLine(client_str, theme, little=True))
                            two_lines_title = True
                    
                    if not two_lines_title:
                        title_lines.append(client_line)
                    
                    subt_len = n_lines - 1
                    if two_lines_title:
                        subt_len -= 1
                        titles = self.split_in_two(subtitle, subt_len)
                        for title in titles:
                            title_lines.append(TitleLine(title, theme))
                    else:
                        titles = self.split_in_two('uuuu' + subtitle, subt_len)
                        
                        # supress the uuuu
                        for i in range(len(titles)):
                            title = titles[i]
                            if i == 0:
                                title = title[4:]
                                if not title:
                                    continue
                            title_lines.append(TitleLine(title, theme))
        else:
            if n_lines >= 2:
                titles = []
                if (self._group_name.startswith(
                     ('Carla.', 'Carla-Multi-Client.', 'Carla-Single-Client.'))
                        and '/' in self._group_name):
                    first_line, slash, last_line = self._group_name.partition('/')
                    titles = [first_line + '/'] + self.split_in_two(last_line, n_lines - 1)
                else:
                    titles = self.split_in_two(self._group_name, n_lines)

                title_lines = [TitleLine(tt, theme) for tt in titles]
            else:
                title_lines = [TitleLine(self._group_name, theme)]

        return tuple(title_lines)
    
    def _choose_title_layout(
        self, height_for_ports: int, height_for_ports_one: int,
        ports_in_width: int, ports_out_width: int) -> dict:
        '''choose in how many lines should be splitted the title
        and if the box layout should be large or high.
        returns a dict with all required variables and values'''
        
        # width_for_ports_one is the width needed for ports
        # if inputs and outputs layouted in one column.
        ## INPUT_PORT_1
        ## INPUT_PORT_2
        ##    OUTPUT_PORT_1
        ## ...
        width_for_ports = 30 + ports_in_width + ports_out_width
        width_for_ports_one = 30 + max(ports_in_width, ports_out_width)

        ports_width = ports_in_width
        if self._current_port_mode is PortMode.OUTPUT:
            ports_width = ports_out_width

        box_theme = self.get_theme()
        font_size = box_theme.font().pixelSize()

        # Check Text Name size
        
        # first look in cache if title sizes are stocked
        all_title_templates = box_theme.get_title_templates(
            self._group_name, self._can_handle_gui, self.has_top_icon())
        lines_choice_max = len(all_title_templates) - 1
        
        if not all_title_templates:
            # this box title is not in cache, we need to estimate all its sizes
            # depending number of splitted lines.
            title_template = {
                "title_width": 0, "header_width": 0, "header_height": 0}
            all_title_templates = [title_template.copy() for i in range(8)]

            last_lines_count = 0
            GUI_MARGIN = 2
            title_line_y_start = 2 + font_size
            
            header_height = 2 + font_size + 2
            if self.has_top_icon():
                header_height = 4 + 24 + 4

            for i in range(1, 8):
                max_title_size = 0
                max_header_width = 50
                if self._plugin_inline != self.INLINE_DISPLAY_DISABLED:
                    max_header_width = 200
                
                title_lines = self._split_title(i)

                for j in range(len(title_lines)):
                    title_line = title_lines[j]
                    title_line.y = title_line_y_start + j * int(font_size * 1.4)
                    max_title_size = int(max(max_title_size, title_line.size))
                    header_width = title_line.size + 12

                    if self.has_top_icon() and title_line.y <= 28 + font_size:
                        # text line is at right of the icon
                        header_width += 28

                    max_header_width = max(max_header_width, header_width)
                
                header_height = max(
                    header_height,
                    title_line_y_start
                        + int(font_size * 1.4) * (len(title_lines) - 1)
                        + 2 + 5)
                
                if self._can_handle_gui:
                    max_header_width += 2 * GUI_MARGIN
                    header_height += 2 * GUI_MARGIN

                new_title_template = title_template.copy()
                new_title_template['title_width'] = max_title_size
                new_title_template['header_width'] = max_header_width
                new_title_template['header_height'] = header_height
                all_title_templates[i] = new_title_template

                if i > 2 and len(title_lines) <= last_lines_count:
                    break

                last_lines_count = len(title_lines)

            lines_choice_max = i
            box_theme.save_title_templates(
                self._group_name, self._can_handle_gui, self.has_top_icon(),
                all_title_templates[:lines_choice_max])

        # Now compare multiple possible areas for the box,
        # depending on BoxLayoutMode and number of lines for box title.

        sizes_tuples = list[tuple[int, int, bool, TitleOn]]()
        layout_mode = self._get_layout_mode_for_this()
        
        if self._current_port_mode in (PortMode.INPUT, PortMode.OUTPUT):
            # splitted box
            
            if layout_mode in (BoxLayoutMode.AUTO, BoxLayoutMode.LARGE):
                ports_y_start_min = box_theme.port_spacing() + box_theme.port_type_spacing()
                
                # calculate area with title on side
                for i in range(1, lines_choice_max + 1):
                    sizes_tuples.append(
                        ((ports_width + all_title_templates[i]['header_width'])
                        * max(all_title_templates[i]['header_height'],
                              height_for_ports + ports_y_start_min),
                        i, False, TitleOn.SIDE))

                if self.has_top_icon():
                    # calculate area with title on side (title under the icon)
                    for i in range(1, lines_choice_max + 1):
                        sizes_tuples.append(
                            ((ports_width + all_title_templates[i]['title_width'] + 16)
                            * max(all_title_templates[i]['header_height'] + 28,
                                  height_for_ports + ports_y_start_min),
                            i, False, TitleOn.SIDE_UNDER_ICON))
            
            if layout_mode in (BoxLayoutMode.AUTO, BoxLayoutMode.HIGH):
                # calculate area with title on top
                for i in range(1, lines_choice_max + 1):
                    sizes_tuples.append(
                        (max(all_title_templates[i]['header_width'], width_for_ports)
                        * (all_title_templates[i]['header_height'] + height_for_ports),
                        i, False, TitleOn.TOP))
        else:
            # --- grouped box ---
            
            # calculate area with input and outputs ports descending
            if (layout_mode in (BoxLayoutMode.AUTO, BoxLayoutMode.HIGH)
                    and options.box_grouped_auto_layout_ratio < 2.0):
                for i in range(1, lines_choice_max + 1):
                    sizes_tuples.append(
                        (max(all_title_templates[i]['header_width'], width_for_ports_one)
                        * (all_title_templates[i]['header_height'] + height_for_ports_one)
                        * options.box_grouped_auto_layout_ratio,
                        i, True, TitleOn.TOP))

            # calculate area with input ports at left of output ports
            if (layout_mode in (BoxLayoutMode.AUTO, BoxLayoutMode.LARGE)
                    and options.box_grouped_auto_layout_ratio > 0.0):
                for i in range(1, lines_choice_max + 1):
                    sizes_tuples.append(
                        (max(all_title_templates[i]['header_width'], width_for_ports)
                        * (all_title_templates[i]['header_height'] + height_for_ports),
                        i, False, TitleOn.TOP))
        
        # sort areas and choose the first one (the littlest area)
        sizes_tuples.sort()
        area_size, lines_choice, one_column, title_on_side = sizes_tuples[0]
        
        self._title_lines = self._split_title(lines_choice)
        
        header_height = all_title_templates[lines_choice]['header_height']
        header_width = all_title_templates[lines_choice]['header_width']
        max_title_size = all_title_templates[lines_choice]['title_width']
        
        if self._current_port_mode is PortMode.BOTH:
            if one_column:
                self._current_layout_mode = BoxLayoutMode.HIGH
            else:
                self._current_layout_mode = BoxLayoutMode.LARGE
        else:
            if title_on_side:
                self._current_layout_mode = BoxLayoutMode.LARGE
            else:
                self._current_layout_mode = BoxLayoutMode.HIGH
        
        box_width = 0
        box_height = 0
        ports_y_start = header_height

        self._title_under_icon = bool(title_on_side is TitleOn.SIDE_UNDER_ICON)

        if title_on_side:
            box_width = ports_width + 12 + header_width
            ports_y_start = box_theme.port_spacing() + box_theme.port_type_spacing()
            box_height = max(height_for_ports + ports_y_start, header_height)
            
            if self._title_under_icon:
                header_width = max(38, max_title_size + 12)
                if self._can_handle_gui:
                    header_width += 4

                box_width = ports_width + header_width + 12
                header_height += 20 if len(self._title_lines) == 1 else 30
                box_height = max(height_for_ports + ports_y_start, header_height)

            ports_y_start = max(ports_y_start, header_height - height_for_ports)

        elif one_column:
            box_width = max(width_for_ports_one, header_width)
            box_height = header_height + height_for_ports_one
        else:
            box_width = max(width_for_ports, header_width)
            box_height = header_height + height_for_ports
        
        return {'max_title_size': max_title_size,
                'box_height': box_height,
                'box_width': box_width,
                'header_width': header_width,
                'header_height': header_height,
                'ports_y_start': ports_y_start,
                'one_column': one_column}
    
    def _set_ports_x_positions(self, max_in_width: int, max_out_width: int):
        box_theme = self.get_theme()
        port_in_offset = box_theme.port_in_offset()
        port_out_offset = box_theme.port_out_offset()
        
        # Horizontal ports re-positioning
        in_x = port_in_offset
        out_x = self._width - max_out_width - 12

        # Horizontal ports not in portgroup re-positioning
        for port in self._port_list:
            if port.portgrp_id:
                continue

            if port.port_mode is PortMode.INPUT:
                port.widget.setX(in_x)
                port.widget.set_port_width(max_in_width - port_in_offset)
            elif port.port_mode is PortMode.OUTPUT:
                port.widget.setX(out_x)
                port.widget.set_port_width(max_out_width - port_out_offset)

        # Horizontal portgroups and ports in portgroup re-positioning
        for portgrp in self._portgrp_list:
            if portgrp.widget is not None:
                if portgrp.port_mode is PortMode.INPUT:
                    portgrp.widget.set_portgrp_width(max_in_width - port_in_offset)
                    portgrp.widget.setX(port_in_offset +1)
                elif portgrp.port_mode is PortMode.OUTPUT:
                    portgrp.widget.set_portgrp_width(max_out_width - port_out_offset)
                    portgrp.widget.setX(out_x)

            max_port_in_pg_width = canvas.theme.port_grouped_width

            for port in portgrp.ports:
                if port.widget is not None:
                    port_print_width = port.widget.get_text_width()

                    # change port in portgroup width only if
                    # portgrp will have a name
                    # to ensure that portgroup widget is large enough
                    max_port_in_pg_width = max(max_port_in_pg_width,
                                               port_print_width + 4)

            out_in_portgrp_x = (self._width - port_out_offset - 12
                                - max_port_in_pg_width)

            portgrp.widget.set_ports_width(max_port_in_pg_width)

            for port in portgrp.ports:
                if port.widget is not None:
                    port.widget.set_port_width(max_port_in_pg_width)
                    if port.port_mode is PortMode.INPUT:
                        port.widget.setX(in_x)
                    elif port.port_mode is PortMode.OUTPUT:
                        port.widget.setX(out_in_portgrp_x)
    
    def _set_title_positions(self):
        ''' set title lines, header lines and icon positions '''
        self._header_line_left = None
        self._header_line_right = None

        box_theme = self.get_theme()
        font_size = box_theme.font().pixelSize()
        font_spacing = int(font_size * 1.4)
        # when client is client capable of gui state
        # header has margins
        gui_margin = 2 if self._can_handle_gui else 0
        
        # set title lines Y position
        title_y_start = font_size + 2 + gui_margin

        if self._title_under_icon:
            title_y_start = 4 + gui_margin + 24 + font_spacing
        
        for i in range(len(self._title_lines)):
            title_line = self._title_lines[i]
            title_line.y = title_y_start + i * font_spacing
        
        if not self._title_under_icon and len(self._title_lines) == 1:
            self._title_lines[0].y = int(
                (self._header_height - font_size) / 2 + font_size -2)
        
        if self._has_side_title():
            y_correction = 0

            # In case the title is near to be vertically centered in the box
            # It's prettier to center it correctly
            if (self._title_lines
                    and not self._title_under_icon
                    and not self._can_handle_gui
                    and self._title_lines[-1].y + int(font_size * 1.4) > self._height):
                y_correction = int((self._height - self._title_lines[-1].y - 2) / 2 - 2)
            
            # set title lines pos
            for title_line in self._title_lines:
                if self._current_port_mode is PortMode.INPUT:
                    title_line.x = 4 + self._width - self._header_width + 2
                    if self._can_handle_gui:
                        title_line.x += 2
                    
                    if self.has_top_icon():
                        self.top_icon.set_pos(self._width - 28 - gui_margin,
                                              4 + gui_margin)
    
                elif self._current_port_mode is PortMode.OUTPUT:
                    title_line.x = (self._header_width
                                    - title_line.size - 6)
                    
                    if self._can_handle_gui:
                        title_line.x -= 2
                    
                    if self.has_top_icon():
                        self.top_icon.set_pos(4 + gui_margin, 4 + gui_margin)
                
                title_line.y += y_correction
            return

        # Now we are sure title is on top

        # get title global sizes
        max_title_size = 0
        max_title_icon_size = 0

        for title_line in self._title_lines:
            title_size = title_line.size
            if self.has_top_icon() and title_line.y <= 28 + font_size:
                # title line is beside icon
                title_size += 28
                max_title_icon_size = max(max_title_icon_size, title_size)
            max_title_size = max(max_title_size, title_size)
        
        # set title lines X position
        for title_line in self._title_lines:
            if self.has_top_icon() and title_line.y <= 28 + font_size:
                # title line is beside the icon
                title_line.x = int((self._width - max_title_icon_size) / 2 + 28)
            else:
                title_line.x = int((self._width - title_line.size) / 2)
        
        # set icon position
        if self.has_top_icon():
            self.top_icon.set_pos(int((self._width - max_title_icon_size)/2),
                                  4 + gui_margin)
        
        # calculate header lines positions
        side_size = int((self._width - max(max_title_icon_size, max_title_size)) / 2)
        
        if side_size > 10:
            y = int(self._header_height / 2)
            
            self._header_line_left = (5, y, side_size - 5, y)
            self._header_line_right = (int(self._width) - side_size + 5, y,
                                       int(self._width) - 5, y)
    
    def build_painter_path(self, pos_dict):
        input_segments = pos_dict['input_segments']
        output_segments = pos_dict['output_segments']
        
        painter_path = QPainterPath()
        theme = self.get_theme()
        border_radius = theme.border_radius()
        port_in_offset = abs(theme.port_in_offset())
        port_out_offset = abs(theme.port_out_offset())
        pen = theme.fill_pen()
        line_hinting = pen.widthF() / 2.0
        
        # theses values are needed to prevent some incorrect painter_path
        # united or subtracted results
        epsy = 0.001
        epsd = epsy * 2.0
        
        rect = QRectF(0.0, 0.0, self._width, self._height)
        rect.adjust(line_hinting, line_hinting, -line_hinting, -line_hinting)
        
        if border_radius == 0.0:
            painter_path.addRect(rect)
        else:
            painter_path.addRoundedRect(rect, border_radius, border_radius)
        
        if not (self._wrapping or self._unwrapping or self._wrapped):
            # substract rects in the box shape in case of port_offset (even negativ)
            # logic would want to add rects if port_offset is negativ
            # But that also means that we should change the boudingRect,
            # So we won't.
            if port_in_offset != 0.0:
                for in_segment in input_segments:
                    moins_path = QPainterPath()
                    moins_path.addRect(QRectF(
                        0.0 - epsy,
                        in_segment[0] - line_hinting - epsy,
                        port_in_offset + line_hinting + epsd,
                        in_segment[1] - in_segment[0] + line_hinting * 2 + epsd))
                    painter_path = painter_path.subtracted(moins_path)
            
            if port_out_offset != 0.0:
                for out_segment in output_segments:
                    moins_path = QPainterPath()
                    moins_path.addRect(QRectF(
                        self._width - line_hinting - port_out_offset - epsy,
                        out_segment[0] - line_hinting - epsy,
                        port_out_offset + line_hinting + epsd,
                        out_segment[1] - out_segment[0] + line_hinting * 2 + epsd))
                    painter_path = painter_path.subtracted(moins_path)

            # No rounded corner if the last port is to close from the corner
            if (input_segments
                    and self._height - input_segments[-1][1] <= border_radius):
                left_path = QPainterPath()
                left_path.addRect(QRectF(
                    0.0 + line_hinting - epsy,
                    max(self._height - border_radius, input_segments[-1][1])
                        + line_hinting - epsy,
                    border_radius + epsd,
                    min(border_radius, self._height - input_segments[-1][1])
                        - 2 * line_hinting + epsd))
                painter_path = painter_path.united(left_path)

            if (input_segments
                    and input_segments[0][0] <= border_radius):
                top_left_path = QPainterPath()
                top_left_path.addRect(QRectF(
                    0.0 + line_hinting - epsy,
                    0.0 + line_hinting - epsy,
                    border_radius + epsd,
                    min(border_radius, input_segments[0][0])
                    - 2 * line_hinting + epsd))
                painter_path = painter_path.united(top_left_path)

            if (output_segments
                    and self._height - output_segments[-1][1] <= border_radius):
                right_path = QPainterPath()
                right_path.addRect(QRectF(
                    self._width - border_radius - line_hinting - epsy,
                    max(self._height - border_radius, output_segments[-1][1])
                        + line_hinting - epsy,
                    border_radius + epsd,
                    min(border_radius, self._height - output_segments[-1][1])
                        - 2 * line_hinting + epsd))
                painter_path = painter_path.united(right_path)
                
            if (output_segments
                    and output_segments[0][0] <= border_radius):
                top_right_path = QPainterPath()
                top_right_path.addRect(QRectF(
                    self._width - line_hinting + epsy - border_radius,
                    0.0 + line_hinting - epsy,
                    border_radius + epsd,
                    min(border_radius, output_segments[0][0])
                    - 2 * line_hinting + epsd))
                painter_path = painter_path.united(top_right_path)
            
        if self.is_monitor() and border_radius:
            if self._current_port_mode is PortMode.OUTPUT:
                left_path = QPainterPath()
                left_path.addRect(QRectF(
                    0.0 + line_hinting - epsy,
                    self._height - border_radius - epsy,
                    border_radius + epsd, border_radius - line_hinting + epsd))
                painter_path = painter_path.united(left_path)

                top_left_path = QPainterPath()
                top_left_path.addRect(QRectF(
                    0.0 + line_hinting - epsy, 0.0 + line_hinting - epsy,
                    border_radius + epsd, border_radius - line_hinting + epsd))
                painter_path = painter_path.united(top_left_path)

            elif self._current_port_mode is PortMode.INPUT:
                right_path = QPainterPath()
                right_path.addRect(QRectF(
                    self._width - line_hinting - epsy - border_radius,
                    self._height - border_radius - epsy,
                    border_radius + epsd, border_radius - line_hinting + epsd))
                painter_path = painter_path.united(right_path)

                top_right_path = QPainterPath()
                top_right_path.addRect(QRectF(
                    self._width - line_hinting - epsy - border_radius,
                    0.0 + line_hinting - epsy,
                    border_radius + epsd, border_radius - line_hinting + epsd))
                painter_path = painter_path.united(top_right_path)

        self._painter_path = painter_path
        
    def update_positions(self, even_animated=False, without_connections=False):
        if canvas.loading_items:
            return

        if (not even_animated
                and self in [b.widget for b in canvas.scene.move_boxes]):
            self.update_positions_pending = True
            # do not change box layout while box is moved by animation
            # update_positions will be called when animation is finished
            return

        self.prepareGeometryChange()
        
        self._current_port_mode = PortMode.NULL
        self._port_list.clear()
        self._portgrp_list.clear()
        
        if self._splitted:
            for port in canvas.list_ports(group_id=self._group_id):
                if port.port_mode is self._splitted_mode:
                    self._port_list.append(port)
            
            if self._port_list:
                self._current_port_mode = self._splitted_mode
        else:
            for port in canvas.list_ports(group_id=self._group_id):
                self._port_list.append(port)
                # used to know present port modes (INPUT or OUTPUT or BOTH)
                self._current_port_mode |= port.port_mode

        for portgrp in canvas.list_portgroups(group_id=self._group_id):
            if self._current_port_mode & portgrp.port_mode:
                self._portgrp_list.append(portgrp)
                
        if options.auto_hide_groups and not self._port_list:
            self.setVisible(False)
            return
        
        self.setVisible(True)
    
        align_port_types = self._should_align_port_types()

        geo_dict = self._get_geometry_dict(align_port_types)
        last_in_pos = geo_dict['last_in_pos']
        last_out_pos = geo_dict['last_out_pos']
        last_inout_pos = geo_dict['last_inout_pos']
        max_in_width = geo_dict['max_in_width']
        max_out_width = geo_dict['max_out_width']
        last_port_mode = geo_dict['last_port_mode']
        
        box_theme = self.get_theme()
        down_height = box_theme.fill_pen().widthF()
        height_for_ports = max(last_in_pos, last_out_pos) + down_height
        height_for_ports_one = last_inout_pos + down_height
       
        self._width_in = max_in_width
        self._width_out = max_out_width

        if not (self._wrapping or self._unwrapping):
            titles_dict = self._choose_title_layout(
                height_for_ports, height_for_ports_one,
                max_in_width, max_out_width)
            self._header_width = titles_dict['header_width']
            self._header_height = titles_dict['header_height']
            one_column = titles_dict['one_column']
            box_width = titles_dict['box_width']
            box_height = titles_dict['box_height']
            self._ports_y_start = titles_dict['ports_y_start']

            self._width = box_width
        
            # wrapped/unwrapped sizes
            normal_height = box_height
            normal_width = box_width
            wrapped_height = self._ports_y_start + canvas.theme.port_height
            wrapped_width = self._width
            
            if self._has_side_title():
                wrapped_height = self._header_height
                if self._current_port_mode is PortMode.INPUT:
                    wrapped_width -= self._width_in
                elif self._current_port_mode is PortMode.OUTPUT:
                    wrapped_width -= self._width_out
            
        else:
            normal_height = self._unwrapped_height
            normal_width = self._unwrapped_width
            wrapped_height = self._wrapped_height
            wrapped_width = self._wrapped_width
            
            one_column = bool(
                self._current_port_mode is PortMode.BOTH
                and self._current_layout_mode == BoxLayoutMode.HIGH)
        
        last_in_pos += self._ports_y_start
        last_out_pos += self._ports_y_start

        if self._wrapping:
            self._height = (normal_height
                            - (normal_height - wrapped_height)
                              * self._wrapping_ratio)
            self._width = (normal_width
                           - (normal_width - wrapped_width)
                             * self._wrapping_ratio)

        elif self._unwrapping:
            self._height = (wrapped_height
                            + (normal_height - wrapped_height)
                              * self._wrapping_ratio)
            self._width = (wrapped_width
                           + (normal_width - wrapped_width)
                             * self._wrapping_ratio)

        elif self._wrapped:
            self._height = wrapped_height
            self._width = wrapped_width
        else:
            self._height = normal_height
            self._width = normal_width
            
            self._unwrap_triangle_pos = UnwrapButton.NONE

            if self._has_side_title():
                if self._height - self._header_height >= 15:
                    if self._current_port_mode is PortMode.OUTPUT:
                        self._unwrap_triangle_pos = UnwrapButton.LEFT
                    else:
                        self._unwrap_triangle_pos = UnwrapButton.RIGHT
            
            if self._height - self._header_height >= 64:
                y_side_space = last_in_pos - last_out_pos
                
                if one_column and last_port_mode is PortMode.INPUT:
                    self._unwrap_triangle_pos = UnwrapButton.RIGHT
                elif one_column and last_port_mode is PortMode.OUTPUT:
                    self._unwrap_triangle_pos = UnwrapButton.LEFT
                elif y_side_space < -10:
                    self._unwrap_triangle_pos = UnwrapButton.LEFT
                elif y_side_space > 10:
                    self._unwrap_triangle_pos = UnwrapButton.RIGHT
                else:
                    self._unwrap_triangle_pos = UnwrapButton.CENTER

        self._wrapped_width = wrapped_width
        self._unwrapped_width = normal_width
        self._wrapped_height = wrapped_height
        self._unwrapped_height = normal_height

        # round self._height to the upper value
        self._height = float(int(self._height + 0.99))
        self._height = ceil(self._height)

        ports_y_segments_dict = self._set_ports_y_positions(
            align_port_types, self._ports_y_start, one_column)

        self._set_ports_x_positions(max_in_width, max_out_width)
        self._set_title_positions()

        if (self._width != self._ex_width
                or self._height != self._ex_height
                or ports_y_segments_dict != self._ex_ports_y_segments_dict):
            self.build_painter_path(ports_y_segments_dict)

        if (self._width != self._ex_width
                or self._height != self._ex_height
                or self.scenePos() != self._ex_scene_pos):
            canvas.scene.resize_the_scene()

        self._ex_width = self._width
        self._ex_height = self._height
        self._ex_ports_y_segments_dict = ports_y_segments_dict
        self._ex_scene_pos = self.scenePos()

        if not without_connections:
            self.repaint_lines(forced=True)

        if not (self._wrapping or self._unwrapping) and self.isVisible():
            canvas.scene.deplace_boxes_from_repulsers([self])

        self.update_positions_pending = False
        self.update()
        
        # I do not understand why without this,
        # when a renamed port does not change the box geometry,
        # port and portgroups widgets aren't updated,
        # they will be if we click on it for exemple.        
        for portgrp in self._portgrp_list:
            if portgrp.widget is not None:
                portgrp.widget.update()
                
        for port in self._port_list:
            if port.widget is not None:
                port.widget.update()



from enum import IntEnum
from PyQt5.QtCore import QObject, pyqtSignal

# we need a QObject for pyqtSignal
class SignalsObject(QObject):
    out_thread_order = pyqtSignal()
    to_main_thread = pyqtSignal(object, tuple, dict)
    callback_sig = pyqtSignal(IntEnum, tuple)
    
    port_types_view_changed = pyqtSignal(int)
    full_screen_toggle_wanted = pyqtSignal()
    filters_bar_toggle_wanted = pyqtSignal()

    theme_changed = pyqtSignal(str)

    # theses signals send int because they are related to checkboxes
    # but the int value is 0 or 1
    graceful_names_changed = pyqtSignal(int)
    a2j_grouped_changed = pyqtSignal(int)
    alsa_midi_enabled_changed = pyqtSignal(int)
    group_shadows_changed = pyqtSignal(int)
    auto_select_items_changed = pyqtSignal(int)
    elastic_changed = pyqtSignal(int)
    borders_nav_changed = pyqtSignal(int)
    prevent_overlap_changed = pyqtSignal(int)
    
    max_port_width_changed = pyqtSignal(int)
    default_zoom_changed = pyqtSignal(int)
    scene_scale_changed = pyqtSignal(float)

    def __init__(self):
        QObject.__init__(self)




import argparse
from asyncio.log import logger
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Optional
try:
    import liblo
except ImportError:
    import pyliblo3 as liblo
import os
import shlex
import socket
import subprocess
import sys
import logging

try:
    from liblo import Server, Address
except ImportError:
    from pyliblo3 import Server, Address
from PyQt5.QtCore import QT_VERSION_STR, QSettings

_logger = logging.getLogger(__name__)

# get qt version in list of ints
QT_VERSION = []
for strdigit in QT_VERSION_STR.split('.'):
    QT_VERSION.append(int(strdigit))

QT_VERSION = tuple(QT_VERSION)

if QT_VERSION < (5, 6):
    sys.stderr.write(
        "WARNING: You are using a version of QT older than 5.6.\n"
        + "You won't be warned if a process can't be launch.\n")

VERSION = "0.14.4"

APP_TITLE = 'RaySession'
DEFAULT_SESSION_ROOT = "%s/Ray Sessions" % os.getenv('HOME')
SCRIPTS_DIR = 'ray-scripts'
NOTES_PATH = 'ray-notes'
FACTORY_SESSION_TEMPLATES = (
    'with_jack_patch', 'with_jack_config', 'scripted')
RAYNET_BIN = 'ray-network'

GIT_IGNORED_EXTENSIONS = ".wav .flac .ogg .mp3 .mp4 .avi .mkv .peak .m4a .pdf"



class PrefixMode:
    CUSTOM = 0
    CLIENT_NAME = 1
    SESSION_NAME = 2


class JackNaming:
    SHORT = 0
    LONG = 1

class ClientStatus:
    STOPPED = 0
    LAUNCH = 1
    OPEN = 2
    READY = 3
    PRECOPY = 4
    COPY = 5
    SAVE = 6
    SWITCH = 7
    QUIT = 8
    NOOP = 9
    ERROR = 10
    REMOVED = 11
    UNDEF = 12
    SCRIPT = 13
    LOSE = 14


class ServerStatus:
    OFF = 0
    NEW = 1
    OPEN = 2
    CLEAR = 3
    SWITCH = 4
    LAUNCH = 5
    PRECOPY = 6
    COPY = 7
    READY = 8
    SAVE = 9
    CLOSE = 10
    SNAPSHOT = 11
    REWIND = 12
    WAIT_USER = 13
    OUT_SAVE = 14
    OUT_SNAPSHOT = 15
    SCRIPT = 16


class NSMMode:
    NO_NSM = 0
    CHILD = 1
    NETWORK = 2


class Protocol:
    NSM = 0
    RAY_HACK = 1
    RAY_NET = 2


class Option:
    NSM_LOCKED = 0x001
    SAVE_FROM_CLIENT = 0x002 #DEPRECATED
    BOOKMARK_SESSION = 0x004
    HAS_WMCTRL = 0x008
    DESKTOPS_MEMORY = 0x010
    HAS_GIT = 0x020
    SNAPSHOTS = 0x040
    SESSION_SCRIPTS = 0x080
    GUI_STATES = 0x100


class Err:
    OK = 0
    GENERAL_ERROR = -1
    INCOMPATIBLE_API = -2
    BLACKLISTED = -3
    LAUNCH_FAILED = -4
    NO_SUCH_FILE = -5
    NO_SESSION_OPEN = -6
    UNSAVED_CHANGES = -7
    NOT_NOW = -8
    BAD_PROJECT = -9
    CREATE_FAILED = -10
    SESSION_LOCKED = -11
    OPERATION_PENDING = -12
    COPY_RUNNING = -13
    NET_ROOT_RUNNING = -14
    SUBPROCESS_UNTERMINATED = -15
    SUBPROCESS_CRASH = -16
    SUBPROCESS_EXITCODE = -17
    UNKNOWN_MESSAGE = -18
    ABORT_ORDERED = -19
    COPY_ABORTED = -20
    SESSION_IN_SESSION_DIR = -21
    # check control/osc_server.py in case of changes !!!


class Command:
    NONE = 0
    START = 1
    OPEN = 2
    SAVE = 3
    STOP = 4


class WaitFor:
    NONE = 0
    QUIT = 1
    STOP_ONE = 2
    ANNOUNCE = 3
    REPLY = 4
    DUPLICATE_START = 5
    DUPLICATE_FINISH = 6
    SCRIPT_QUIT = 7


class Template(IntEnum):
    NONE = 0
    RENAME = 1
    SESSION_SAVE = 2
    SESSION_SAVE_NET = 3
    SESSION_LOAD = 4
    SESSION_LOAD_NET = 5
    CLIENT_SAVE = 6
    CLIENT_LOAD = 7


class SwitchState:
    NONE = 0
    RESERVED = 1
    NEEDED = 2
    DONE = 3


class WindowManager:
    NONE = 0
    X = 1
    WAYLAND = 2


class Systray:
    OFF = 0
    SESSION_ONLY = 1
    ALWAYS = 2

@dataclass
class Favorite:
    name: str
    icon: str
    factory: bool
    display_name: str


class ScriptFile:
    PREVENT = 0x0
    PARENT = 0x1
    LOAD = 0x2
    SAVE = 0x4
    CLOSE = 0x8

    @classmethod
    def by_string(cls, action: str) -> int:
        if action == 'load':
            return cls.LOAD
        if action == 'save':
            return cls.SAVE
        if action == 'close':
            return cls.CLOSE
        return cls.PREVENT

debug = False


def ifDebug(string):
    if debug:
        sys.stderr.write("%s\n" % string)

def version_to_tuple(version_str: str):
    version_list = []
    for c in version_str.split('.'):
        if not c.isdigit():
            return ()
        version_list.append(int(c))

    return tuple(version_list)

def add_self_bin_to_path():
    # Add RaySession/src/bin to $PATH to can use ray executables after make
    # Warning, will works only if link to this file is in RaySession/*/*/*.py
    this_path = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))
    bin_path = "%s/bin" % os.path.dirname(this_path)
    if not os.environ['PATH'].startswith("%s:" % bin_path):
        os.environ['PATH'] = "%s:%s" % (bin_path, os.environ['PATH'])

def get_list_in_settings(settings: QSettings, path: str) -> list:
    # getting a QSettings value of list type seems to not works the same way
    # on all machines
    try:
        settings_list = settings.value(path, [], type=list)
    except BaseException:
        try:
            settings_list = settings.value(path, [])
        except BaseException:
            settings_list = []

    if isinstance(settings_list, list):
        return settings_list
    return []

def is_git_taggable(string: str) -> bool:
    ''' know if a string can be a git tag, not used currently '''
    if not string:
        return False

    if string.startswith('/'):
        return False

    if string.endswith('/'):
        return False

    if string.endswith('.'):
        return False

    for forbidden in (' ', '~', '^', ':', '?', '*',
                      '[', '..', '@{', '\\', '//', ','):
        if forbidden in string:
            return False

    if string == "@":
        return False

    return True

def is_valid_full_path(path: str) -> bool:
    if not path.startswith('/'):
        return False

    for forbidden in ('//', '/./', '/../'):
        if forbidden in path:
            return False

    if path.endswith(('/.', '/..')):
        return False
    return True

def is_osc_port_free(port: int) -> bool:
    try:
        testport = Server(port)
    except BaseException:
        return False

    del testport
    return True


def get_free_osc_port(default=16187):
    ''' get a free OSC port for daemon, start from default '''

    if default >= 65536:
        default = 16187

    daemon_port = default
    used_port = True
    testport = None

    while used_port:
        try:
            testport = Server(daemon_port)
            used_port = False
        except BaseException:
            daemon_port += 1
            used_port = True

    del testport
    return daemon_port

def is_valid_osc_url(url: str) -> bool:
    try:
        address = liblo.Address(url)
        return True
    except BaseException:
        return False

def get_liblo_address(url: str) -> Optional[liblo.Address]:
    valid_url = False
    try:
        address = liblo.Address(url)
        valid_url = True
    except BaseException:
        valid_url = False
        msg = "%r is not a valid osc url" % url
        raise argparse.ArgumentTypeError(msg)

    if valid_url:
        try:
            liblo.send(address, '/ping')
            return address
        except BaseException:
            msg = "%r is an unknown osc url" % url
            raise argparse.ArgumentTypeError(msg)

def get_liblo_address_from_port(port:int) -> Optional[liblo.Address]:
    try:
        port = int(port)
    except:
        msg = "%r port must be an int" % port
        raise argparse.ArgumentTypeError(msg)

    valid_port = False

    try:
        address = liblo.Address(port)
        valid_port = True
    except BaseException:
        valid_port = False
        msg = "%i is not a valid osc port" % port
        raise argparse.ArgumentTypeError(msg)

    if valid_port:
        try:
            liblo.send(address, '/ping')
            return address
        except BaseException:
            msg = "%i is an unknown osc port" % port
            raise argparse.ArgumentTypeError(msg)

def are_same_osc_port(url1, url2):
    if url1 == url2:
        return True
    try:
        address1 = Address(url1)
        address2 = Address(url2)
    except BaseException:
        return False

    if address1.port != address2.port:
        return False

    if are_on_same_machine(url1, url2):
        return True

    return False

def are_on_same_machine(url1, url2):
    if url1 == url2:
        return True

    try:
        address1 = Address(url1)
        address2 = Address(url2)
    except BaseException:
        return False

    if address1.hostname == address2.hostname:
        return True

    try:
        if (socket.gethostbyname(address1.hostname)
                    in ('127.0.0.1', '127.0.1.1')
                and socket.gethostbyname(address2.hostname)
                    in ('127.0.0.1', '127.0.1.1')):
            return True

        if socket.gethostbyaddr(
                address1.hostname) == socket.gethostbyaddr(
                address2.hostname):
            return True

        ip = Machine192.get()

        if ip not in (address1.hostname, address2.hostname):
            return False

        if (ip == socket.gethostbyname(address1.hostname)
                == socket.gethostbyname(address2.hostname)):
            # on some systems (as fedora),
            # socket.gethostbyname returns a 192.168.. url
            return True

        if (socket.gethostbyname(address1.hostname)
                in ('127.0.0.1', '127.0.1.1')):
            if address2.hostname == ip:
                return True

        if (socket.gethostbyname(address2.hostname)
                in ('127.0.0.1', '127.0.1.1')):
            if address1.hostname == ip:
                return True

    except BaseException:
        return False

    return False

def get_net_url(port) -> str:
    ip = Machine192.get()
    if not ip:
        return ''

    return "osc.udp://%s:%i/" % (ip, port)

def shell_line_to_args(string:str) -> list:
    try:
        args = shlex.split(string)
    except BaseException:
        return None

    return args

def types_are_all_strings(types:str) -> bool:
    for char in types:
        if char != 's':
            return False
    return True

def are_they_all_strings(args:list) -> bool:
    for arg in args:
        if not isinstance(arg, str):
            return False
    return True

def get_window_manager() -> WindowManager:
    if os.getenv('WAYLAND_DISPLAY'):
        return WindowManager.WAYLAND

    if os.getenv('DISPLAY'):
        return WindowManager.X

    return WindowManager.NONE

def protocol_to_str(protocol: int) -> str:
    if protocol == Protocol.RAY_HACK:
        return "Ray-Hack"
    if protocol == Protocol.RAY_NET:
        return "Ray-Net"
    return "NSM"

def protocol_from_str(protocol_str: str) -> int:
    if protocol_str.lower() in ('ray_hack', 'ray-hack'):
        return Protocol.RAY_HACK
    elif protocol_str.lower() in ('ray_net', 'ray-net'):
        return Protocol.RAY_NET
    return Protocol.NSM


class Machine192:
    ip = ''
    read_done = False
    
    @staticmethod
    def read() -> str:
        try:
            ips = subprocess.check_output(
                ['ip', 'route', 'get', '1']).decode()
            ip_line = ips.partition('\n')[0]
            ip_end = ip_line.rpartition('src ')[2]
            ip = ip_end.partition(' ')[0]

        except BaseException:
            try:
                ips = subprocess.check_output(['hostname', '-I']).decode()
                ip = ips.split(' ')[0]
            except BaseException:
                return ''

        if ip.count('.') != 3:
            return ''
        
        return ip
    
    @classmethod
    def get(cls)->str:
        if cls.read_done:
            return cls.ip
        
        cls.ip = cls.read()
        cls.read_done = True

        return cls.ip


class ClientData:
    client_id = ''
    protocol = Protocol.NSM
    executable_path = ''
    arguments = ''
    pre_env = ''
    name = ''
    prefix_mode = PrefixMode.SESSION_NAME
    custom_prefix = ''
    desktop_file = ''
    label = ''
    description = ''
    icon = ''
    capabilities = ''
    check_last_save = True
    ignored_extensions = GIT_IGNORED_EXTENSIONS
    template_origin = ''
    jack_client_name = ''
    jack_naming = 0
    in_terminal = False
    ray_hack: 'RayHack' = None
    ray_net: 'RayNet' = None

    @staticmethod
    def sisi():
        return 'sissssissssssisssii'

    @staticmethod
    def new_from(*args):
        client_data = ClientData()
        client_data.update(*args)
        return client_data

    @staticmethod
    def spread_client(client: 'ClientData') -> tuple:
        return (client.client_id, client.protocol,
                client.executable_path, client.arguments, client.pre_env,
                client.name, client.prefix_mode, client.custom_prefix,
                client.desktop_file, client.label, client.description,
                client.icon,
                client.capabilities, int(client.check_last_save),
                client.ignored_extensions,
                client.template_origin,
                client.jack_client_name, client.jack_naming,
                int(client.in_terminal))

    def set_ray_hack(self, ray_hack):
        self.ray_hack = ray_hack

    def set_ray_net(self, ray_net):
        self.ray_net = ray_net

    def update(self, client_id, protocol,
               executable, arguments, pre_env,
               name, prefix_mode, custom_prefix,
               desktop_file, label, description,
               icon,
               capabilities, check_last_save,
               ignored_extensions,
               template_origin,
               jack_client_name, jack_naming,
               in_terminal,
               secure=False):
        self.executable_path = str(executable)
        self.arguments = str(arguments)
        self.pre_env = str(pre_env)

        self.desktop_file = str(desktop_file)
        self.label = str(label)
        self.description = str(description)
        self.icon = str(icon)

        self.check_last_save = bool(check_last_save)
        self.ignored_extensions = str(ignored_extensions)
        self.template_origin = template_origin
        self.jack_naming = jack_naming
        self.in_terminal = bool(in_terminal)

        if secure:
            return

        # Now, if message is 'unsecure' only.
        # change things that can't be changed normally
        self.client_id = str(client_id)
        self.protocol = int(protocol)
        if name:
            self.name = str(name)
        else:
            self.name = os.path.basename(self.executable_path)
        self.prefix_mode = int(prefix_mode)

        if self.prefix_mode == PrefixMode.CUSTOM:
            if custom_prefix:
                self.custom_prefix = str(custom_prefix)
            else:
                self.prefix_mode = PrefixMode.SESSION_NAME

        self.capabilities = str(capabilities)
        self.jack_client_name = jack_client_name

    def update_secure(self, *args):
        self.update(*args, secure=True)

    def spread(self) -> tuple:
        return ClientData.spread_client(self)
    
    def prettier_name(self) -> str:
        if self.label:
            return self.label
        if (self.protocol != Protocol.RAY_HACK
                and self.name):
            return self.name
        return self.executable_path


class RayHack:
    config_file = ""
    save_sig = 0
    stop_sig = 15
    wait_win = False
    no_save_level = 0
    useless_str = ''
    useless_int = 0

    @staticmethod
    def sisi():
        # the first 's' is for client_id, not stocked in RayHack
        return 'siiiisi'

    @staticmethod
    def new_from(*args):
        ray_hack = RayHack()
        ray_hack.update(*args)
        return ray_hack

    def saveable(self)->bool:
        return bool(self.config_file and self.save_sig)

    def noSaveLevel(self)->int:
        if self.config_file and self.save_sig == 0:
            return self.no_save_level

        return 0

    def update(self, config_file,
               save_sig, stop_sig,
               wait_win, no_save_level,
               useless_str, useless_int):
        self.config_file = str(config_file)
        self.save_sig = int(save_sig)
        self.stop_sig = int(stop_sig)
        self.wait_win = bool(wait_win)
        self.no_save_level = int(no_save_level)

    def spread(self)->tuple:
        return (self.config_file, self.save_sig, self.stop_sig,
                int(self.wait_win), self.no_save_level,
                self.useless_str, self.useless_int)


class RayNet:
    daemon_url = ''
    session_root = ''
    session_template = ''
    duplicate_state = -1
    running_daemon_url = ''
    running_session_root =''

    @staticmethod
    def sisi():
        return 'sss'

    @staticmethod
    def new_from(*args):
        ray_net = RayNet()
        ray_net.update(*args)
        return ray_net

    def update(self, daemon_url, session_root, session_template):
        self.daemon_url = daemon_url
        self.session_root = session_root
        self.session_template = session_template

    def spread(self)->tuple:
        return (self.daemon_url, self.session_root, self.session_template)


class GroupPosition:
    port_types_view = 3
    group_name = ''
    null_zone = ''
    in_zone = ''
    out_zone = ''
    null_xy = (0, 0)
    in_xy = (0, 0)
    out_xy = (0, 0)
    flags = 0
    layout_mode = 0
    fully_set = True
    
    @staticmethod
    def get_attributes():
        return ('port_types_view', 'group_name',
                'null_zone', 'in_zone', 'out_zone',
                'null_xy', 'in_xy', 'out_xy', 'flags',
                'layout_mode')
    
    @staticmethod
    def sisi():
        return 'issssiiiiiiii'
    
    @staticmethod
    def new_from(*args):
        group_position = GroupPosition()
        group_position.update(*args)
        return group_position
    
    def write_from_dict(self, input_dict: dict):
        for attr in input_dict:
            if not attr in self.get_attributes():
                sys.stderr.write(
                    'group position has no attribute %s\n' % attr)
                continue
            
            value = input_dict[attr]
            attr_type = type(value)
            if attr in ('port_types_view', 'flags'):
                if attr_type != int:
                    continue
            elif attr in ('group_name', 'null_zone', 'in_zone', 'out_zone'):
                if attr_type != str:
                    continue
            elif attr in ('null_xy', 'in_xy', 'out_xy'):
                if attr_type not in (list, tuple):
                    continue
                value = tuple(value)
            
            self.__setattr__(attr, value)
    
    def is_same(self, other: 'GroupPosition') -> bool:
        if (self.port_types_view == other.port_types_view
                and self.group_name == other.group_name):
            return True
        
        return False
    
    def update(self, port_types_view: int, group_name: str,
               null_zone: str, in_zone: str, out_zone: str,
               null_x: int, null_y: int, in_x: int, in_y: int,
               out_x: int, out_y: int, flags: int, layout_mode: int):
        
        for string in (group_name, null_zone, in_zone, out_zone):
            if not isinstance(string, str):
                return 
        
        for digit in (port_types_view, null_x, null_y, in_x, in_y,
                      out_x, out_y, flags, layout_mode):
            if isinstance(digit, int):
                continue

            if not isinstance(digit, str):
                return
            
            if (digit.isdigit()
                    or (digit.startswith('-')
                        and digit.replace('-', '', 1).isdigit())):
                continue
            else:
                return

        self.port_types_view = port_types_view
        self.group_name = group_name
        self.null_zone = null_zone
        self.in_zone = in_zone
        self.out_zone = out_zone
        self.null_xy = (int(null_x), int(null_y))
        self.in_xy = (int(in_x), int(in_y))
        self.out_xy = (int(out_x), int(out_y))
        self.flags = int(flags)
        self.layout_mode = int(layout_mode)
        
    def spread(self) -> tuple:
        return (self.port_types_view, self.group_name,
                self.null_zone, self.in_zone, self.out_zone,
                self.null_xy[0], self.null_xy[1], self.in_xy[0], self.in_xy[1],
                self.out_xy[0], self.out_xy[1], self.flags,
                self.layout_mode)
    
    def to_dict(self) -> dict:
        new_dict = {}
        
        for attr in self.__dir__():
            if attr in self.get_attributes():
                new_dict[attr] = self.__getattribute__(attr)
        
        return new_dict
    
    def get_str_value(self, attr: str)->str:
        if attr not in self.get_attributes():
            return ''
        
        return str(self.__getattribute__(attr))
    
    def set_layout_mode(self, port_mode: int, layout_mode: int):
        if not (1 <= port_mode <= 3 or 0 <= layout_mode <= 2):
            logger.warning(
                "group position set_layout_mode, wrong port mode or layout mode")
            return

        layout_mode_str = "%03d" % self.layout_mode
        new_string = ''
        for i in range(len(layout_mode_str)):
            if i == 3 - port_mode:
                new_string += str(layout_mode)
            else:
                new_string += layout_mode_str[i]
        
        self.layout_mode = int(new_string)
        
    def get_layout_mode(self, port_mode: int) -> int:
        if not(1 <= port_mode <= 3):
            return 0
        
        layout_mode_str = "%03d" % self.layout_mode
        return int(layout_mode_str[3 - port_mode])


class PortGroupMemory:
    group_name = ''
    port_type = 0
    port_mode = 0
    port_names = list[str]()
    above_metadatas = False
    
    @staticmethod
    def get_attributes():
        return ('group_name', 'port_type', 'port_mode',
                'above_metadatas', 'port_names')
    
    @staticmethod
    def new_from(*args):
        portgrp_memory = PortGroupMemory()
        portgrp_memory.update(*args)
        return portgrp_memory
    
    def write_from_dict(self, input_dict: dict):
        for attr in input_dict:
            if not attr in self.get_attributes():
                sys.stderr.write(
                    'PortGroupMemory has no attribute %s\n' % attr)
                continue

            value = input_dict[attr]
            attr_type = type(value)
            if value == 'group_name' and attr_type != str:
                continue
            if value in ('port_type', 'port_mode') and attr_type != int:
                continue
            if value == 'above_metadatas' and attr_type != bool:
                continue
            if value == 'port_names' and attr_type not in (tuple, list):
                continue

            self.__setattr__(attr, value)
    
    def update(self, group_name: str, port_type: int, port_mode: int,
               above_metadatas: int, *port_names):
        self.group_name = group_name
        self.port_type = port_type
        self.port_mode = port_mode
        self.above_metadatas = bool(above_metadatas)
        self.port_names = port_names
    
    def spread(self)->tuple:
        return (self.group_name, self.port_type, self.port_mode,
                int(self.above_metadatas), *self.port_names)
    
    def to_dict(self)->dict:
        new_dict = {}
        
        for attr in self.__dir__():
            if attr in self.get_attributes():
                new_dict[attr] = self.__getattribute__(attr)
        
        return new_dict
    
    def has_a_common_port_with(self, other: 'PortGroupMemory') -> bool:
        if (self.port_type != other.port_type
                or self.port_mode != other.port_mode
                or self.group_name != other.group_name):
            return False
        
        for port_name in self.port_names:
            if port_name in other.port_names:
                return True
        
        return False




import json
import os
import tempfile
import time
from typing import TYPE_CHECKING
try:
    from liblo import Address
except ImportError:
    from pyliblo3 import Address

import ray

from daemon_tools import RS, dirname, Terminal
from server_sender import ServerSender
from jack_renaming_tools import group_belongs_to_client

if TYPE_CHECKING:
    from session_signaled import SignaledSession

JSON_PATH = 'ray_canvas.json'

def _get_version_tuple_json_dict(json_contents: dict) -> tuple[int, int, int]:
    if 'version' in json_contents.keys():
        version_str = json_contents['version']
        try:
            version_list = [int(v) for v in version_str.split('.')]
        except:
            version_list = [int(v) for v in ray.VERSION.split('.')]
    else:
        version_list = [0, 12, 0]
    
    return tuple(version_list)


class CanvasSaver(ServerSender):
    def __init__(self, session: 'SignaledSession'):
        ServerSender.__init__(self)
        self.session = session

        self.group_positions_session = list[ray.GroupPosition]()
        self.group_positions_config = list[ray.GroupPosition]()
        self.portgroups = list[ray.PortGroupMemory]()
        self._config_json_path = "%s/%s" % (
            dirname(RS.settings.fileName()), JSON_PATH)

        if not os.path.exists(self._config_json_path):
            return

        with open(self._config_json_path, 'r') as f:
            json_contents = {}
            gpos_list = list[ray.GroupPosition]()
            pg_list = list[ray.PortGroupMemory]()

            try:
                json_contents = json.load(f)
            except json.JSONDecodeError:
                Terminal.message(f"Failed to load patchcanvas config file {f}")

            # Old group_position port_types_view norm was
            # full_view = AUDIO | MIDI (1|2 = 3)
            # now this is:
            # full_view = AUDIO | MIDI | CV | VIDEO (1|2|4|8 = 15)
            # (even if VIDEO is not implemented in patchbay yet)
            
            # so we need to consider that a full view is a full view
            # and convert in old sessions port_types_view = 3 -> 15 
            needs_port_types_view_convert = False

            if isinstance(json_contents, dict):
                if 'group_positions' in json_contents.keys():
                    gpos_list = json_contents['group_positions']
                if 'portgroups' in json_contents.keys():
                    pg_list = json_contents['portgroups']
                    
                if _get_version_tuple_json_dict(json_contents) < (0, 13, 0):
                    needs_port_types_view_convert = True

            for gpos_dict in gpos_list:
                gpos = ray.GroupPosition()
                gpos.write_from_dict(gpos_dict)
                if needs_port_types_view_convert and gpos.port_types_view == 3:
                    gpos.port_types_view = 15
                
                if not [g for g in self.group_positions_config if g.is_same(gpos)]:
                    self.group_positions_config.append(gpos)

            for pg_dict in pg_list:
                portgroup = ray.PortGroupMemory()
                portgroup.write_from_dict(pg_dict)
                self.portgroups.append(portgroup)

    def get_all_group_positions(self) -> list[ray.GroupPosition]:
        group_positions_config_exclu = list[ray.GroupPosition]()

        for gpos_cf in self.group_positions_config:
            for gpos_ss in self.group_positions_session:
                if (gpos_ss.port_types_view == gpos_cf.port_types_view
                        and gpos_ss.group_name == gpos_cf.group_name):
                    break
            else:
                group_positions_config_exclu.append(gpos_cf)

        return self.group_positions_session + group_positions_config_exclu

    def send_session_group_positions(self):
        server = self.get_server()
        if not server:
            return

        local_guis = []
        distant_guis = []
        for gui_addr in server.gui_list:
            if ray.are_on_same_machine(server.url, gui_addr.url):
                local_guis.append(gui_addr)
            else:
                distant_guis.append(gui_addr)

        if local_guis:
            session_gpos_dict = {'group_positions': list[dict]()}
            for gpos in self.group_positions_session:
                session_gpos_dict['group_positions'].append(gpos.to_dict())

            for gui_addr in local_guis:
                file = tempfile.NamedTemporaryFile(delete=False, mode='w+')
                json.dump(session_gpos_dict, file)

                self.send(gui_addr,
                          '/ray/gui/patchbay/fast_temp_file_memory',
                          file.name)

        if distant_guis:
            for gui_addr in distant_guis:
                for gpos in self.group_positions_session:
                    self.send(gui_addr,
                              '/ray/gui/patchbay/update_group_position',
                              *gpos.spread())

    def send_all_group_positions(self, src_addr: Address):
        if ray.are_on_same_machine(self.get_server_url(), src_addr.url):
            # GUI is on the same machine than the daemon
            # send group positions via a tmp file because they can be many
            # it can be faster, it also prevents to lose packets
            canvas_dict = {'group_positions': [], 'portgroups': []}
            for gpos in self.group_positions_session:
                canvas_dict['group_positions'].append(gpos.to_dict())

            for gpos_cf in self.group_positions_config:
                for gpos_ss in self.group_positions_session:
                    if (gpos_ss.port_types_view == gpos_cf.port_types_view
                            and gpos_ss.group_name == gpos_cf.group_name):
                        break
                else:
                    canvas_dict['group_positions'].append(gpos_cf.to_dict())

            for portgroup in self.portgroups:
                canvas_dict['portgroups'].append(portgroup.to_dict())

            file = tempfile.NamedTemporaryFile(delete=False, mode='w+')
            json.dump(canvas_dict, file)
            file.close()
            self.send(src_addr,
                      '/ray/gui/patchbay/fast_temp_file_memory',
                      file.name)
            return

        i = 0

        for gpos in self.group_positions_session:
            self.send(src_addr, '/ray/gui/patchbay/update_group_position',
                      *gpos.spread())
            i += 1
            if i == 50:
                # we need to slow big process of canvas memory
                # to prevent loss OSC packets
                time.sleep(0.020)
                i = 0

        for gpos_cf in self.group_positions_config:
            for gpos_ss in self.group_positions_session:
                if (gpos_ss.port_types_view == gpos_cf.port_types_view
                        and gpos_ss.group_name == gpos_cf.group_name):
                    break
            else:
                self.send(src_addr, '/ray/gui/patchbay/update_group_position',
                          *gpos_cf.spread())

                i += 1
                if i == 50:
                    time.sleep(0.020)
                    i = 0

        for portgroup in self.portgroups:
            self.send(src_addr, '/ray/gui/patchbay/update_portgroup',
                      *portgroup.spread())

            i += 1
            if i == 50:
                time.sleep(0.020)

    def save_group_position(self, *args):
        gp = ray.GroupPosition.new_from(*args)
        for group_positions in (self.group_positions_session,
                                self.group_positions_config):
            for gpos in group_positions:
                if gpos.is_same(gp):
                    gpos.update(*args)
                    break
            else:
                group_positions.append(gp)

    def load_json_session_canvas(self, session_path: str):
        self.group_positions_session.clear()

        session_canvas_file = "%s/.%s" % (session_path, JSON_PATH)

        if not os.path.exists(session_canvas_file):
            return

        with open(session_canvas_file, 'r') as f:
            json_contents = {}
            gpos_list = list[dict]()

            try:
                json_contents = json.load(f)
            except json.JSONDecodeError:
                Terminal.message("Failed to load session canvas file %s" % f)

            session_version = (0, 14, 0)

            if isinstance(json_contents, dict):
                if 'group_positions' in json_contents.keys():
                    gpos_list : list[dict]() = json_contents['group_positions']
                session_version = _get_version_tuple_json_dict(json_contents)

            for gpos_dict in gpos_list:
                gpos = ray.GroupPosition()
                gpos.write_from_dict(gpos_dict)
                if session_version < (0, 13, 0):
                    if gpos.port_types_view == 3:
                        gpos.port_types_view = 15
                
                elif session_version < (0, 14, 0):
                    if gpos.port_types_view == 15:
                        gpos.port_types_view = 31
                
                if not [g for g in self.group_positions_session if g.is_same(gpos)]:
                    self.group_positions_session.append(gpos)

    def save_json_session_canvas(self, session_path: str):
        session_json_path = "%s/.%s" % (session_path, JSON_PATH)

        if not self.group_positions_session:
            return

        json_contents = {}
        json_contents['group_positions'] = [
            gpos.to_dict() for gpos in self.group_positions_session]
        json_contents['version'] = ray.VERSION

        with open(session_json_path, 'w+') as f:
            json.dump(json_contents, f, indent=2)

    def save_config_file(self):
        if not self.group_positions_config:
            return

        json_contents = {}
        json_contents['group_positions'] = [
            gpos.to_dict() for gpos in self.group_positions_config]
        json_contents['portgroups'] = [
            portgroup.to_dict() for portgroup in self.portgroups]
        json_contents['version'] = ray.VERSION

        with open(self._config_json_path, 'w+') as f:
            json.dump(json_contents, f, indent=2)

    def save_portgroup(self, *args):
        new_portgroup = ray.PortGroupMemory.new_from(*args)

        remove_list = []

        # remove any portgroup with a commmon port with the new one
        for portgroup in self.portgroups:
            if portgroup.has_a_common_port_with(new_portgroup):
                remove_list.append(portgroup)

        for portgroup in remove_list:
            self.portgroups.remove(portgroup)

        self.portgroups.append(new_portgroup)

    def client_jack_name_changed(self, old_jack_name: str, new_jack_name: str):
        server = self.session.get_server()
        
        for gpos in self.group_positions_session:
            if group_belongs_to_client(gpos.group_name, old_jack_name):
                gpos.group_name = gpos.group_name.replace(
                    old_jack_name, new_jack_name, 1)

                if server is not None:
                    server.send_gui(
                        '/ray/gui/patchbay/update_group_position',
                        *gpos.spread())



import enum
from PyQt5.QtCore import QObject, pyqtSignal
try:
    from liblo import Address
except ImportError:
    from pyliblo3 import Address

class Signaler(QObject):
    osc_receive = pyqtSignal(str, list)
    daemon_announce = pyqtSignal(Address, str, int, int, str, int)
    daemon_announce_ok = pyqtSignal()
    daemon_nsm_locked = pyqtSignal(bool)
    server_copying = pyqtSignal(bool)

    add_sessions_to_list = pyqtSignal(list)
    new_executable = pyqtSignal(list)
    session_template_found = pyqtSignal(list)
    user_client_template_found = pyqtSignal(list)
    factory_client_template_found = pyqtSignal(list)
    snapshots_found = pyqtSignal(list)
    reply_auto_snapshot = pyqtSignal(bool)
    server_progress = pyqtSignal(float)
    client_progress = pyqtSignal(str, float)
    server_status_changed = pyqtSignal(int)

    daemon_url_request = pyqtSignal(int, str)
    daemon_url_changed = pyqtSignal(str)

    client_template_update = pyqtSignal(list)
    client_template_ray_hack_update = pyqtSignal(list)
    client_template_ray_net_update = pyqtSignal(list)

    root_changed = pyqtSignal(str)

    session_preview_update = pyqtSignal()
    session_details = pyqtSignal(str, int, int, int)
    scripted_dir = pyqtSignal(str, int)
    parrallel_copy_state = pyqtSignal(int, int)
    parrallel_copy_progress = pyqtSignal(int, float)
    parrallel_copy_aborted = pyqtSignal()
    other_session_renamed = pyqtSignal()
    other_session_duplicated = pyqtSignal()
    other_session_templated = pyqtSignal()

    client_added_reply = pyqtSignal(str)

    client_properties_state_changed = pyqtSignal(str, bool)

    favorite_added = pyqtSignal(str, str, bool, str)
    favorite_removed = pyqtSignal(str, bool)

    hiddens_changed = pyqtSignal(int)

    canvas_callback = pyqtSignal(enum.IntEnum, tuple)

    def __init__(self):
        QObject.__init__(self)
